#include "RaveBase/RaveInterface/rave/KinematicTreeFactory.h"

#include "FWCore/MessageLogger/interface/MessageLogger.h"
#include "RecoVertex/KinematicFit/interface/KinematicParticleVertexFitter.h"
#include "RecoVertex/KinematicFit/interface/KinematicConstrainedVertexFitter.h"
#include "RecoVertex/KinematicFit/interface/KinematicParticleFitter.h"
#include "RecoVertex/KinematicFit/interface/LagrangeParentParticleFitter.h"
#include "RecoVertex/KinematicFit/interface/LagrangeChildUpdator.h"
#include "RecoVertex/KinematicFitPrimitives/interface/KinematicConstraint.h"
#include "RecoVertex/KinematicFitPrimitives/interface/MultiTrackKinematicConstraint.h"

#include "RaveBase/Converters/interface/MagneticFieldWrapper.h"
#include "RaveBase/RaveEngine/interface/RaveBeamSpotSingleton.h"
#include "RaveBase/RaveEngine/interface/RaveParameterSetBuilder.h"
#include "RaveTools/Converters/interface/MagneticFieldSingleton.h"

#include "boost/bind.hpp"

namespace rave
{

KinematicTreeFactory::KinematicTreeFactory (
  const rave::MagneticField & f, const rave::Propagator & p, int verbosity ) :
    theField ( f.copy() ), thePropagator ( p.copy() ),
    theVerbosity ( verbosity )
{
  setup();
}

rave::KinematicTreeFactory::KinematicTreeFactory (
  const rave::MagneticField & f, const rave::Propagator & p,
  const rave::Ellipsoid3D & beamspot, int verbosity ) :
    theField ( f.copy() ), thePropagator ( p.copy() ),
    theVerbosity ( verbosity )
{
  rave::BeamSpotSingleton::set ( beamspot );
  setup();
}

KinematicTreeFactory::~KinematicTreeFactory()
{}

void KinematicTreeFactory::setup()
{
  MagneticFieldSingleton::Instance()->registry ( new MagneticFieldWrapper ( *theField ) );

  edm::LogInfo ( "rave::KinematicTreeFactory" )
  << " magnetic field at (0,0,0) is "
  << MagneticFieldSingleton::Instance()->inTesla ( GlobalPoint ( 0., 0., 0. ) )
  << endl;
}

namespace
{
::RefCountedKinematicParticle doParticleCasting ( const boost::any & p )
{
  return boost::any_cast< ::RefCountedKinematicParticle > ( p );
}

::RefCountedKinematicTree doTreeCasting ( const boost::any & t )
{
  return boost::any_cast< ::RefCountedKinematicTree > ( t );
}
}

KinematicTree KinematicTreeFactory::useVertexFitter (
  std::vector< KinematicParticle > particles, std::string parameters ) const
{
  try {
    // The fitter we are going to use
    ::KinematicParticleVertexFitter fitter;

    // Create the input data field
    std::vector< ::RefCountedKinematicParticle > internalParticles;
    std::transform (
      particles.begin(),
      particles.end(),
      std::back_inserter ( internalParticles ),
      boost::bind ( &doParticleCasting,
                    boost::bind ( &BasicKinematicParticle::getInternal,
                                  boost::bind ( &KinematicParticle::data, _1 ) ) ) );

    // Do the fitting, this is where the real work is done
    ::RefCountedKinematicTree internalTree = fitter.fit ( internalParticles );

    // Now parse the tree into the interface format
    return rave::KinematicTree ( BasicKinematicTree ( internalTree, particles ) );
  } catch ( VertexException e ) {
    edm::LogError ( "rave::KinematicTreeFactory" ) << "VertexException saying """ << e.what() << """ occured.";
    return rave::KinematicTree();
  }
}

rave::KinematicTree KinematicTreeFactory::useVertexFitter (
  std::vector< rave::KinematicParticle > particles,
  rave::KinematicConstraint cs, std::string parameters ) const
{
  try {
    // The fitter we are going to use
    ::KinematicConstrainedVertexFitter fitter;

    // Create the input data field
    std::vector< ::RefCountedKinematicParticle > internalParticles;
    std::transform (
      particles.begin(),
      particles.end(),
      std::back_inserter ( internalParticles ),
      boost::bind ( &doParticleCasting,
                    boost::bind ( &BasicKinematicParticle::getInternal,
                                  boost::bind ( &KinematicParticle::data, _1 ) ) ) );

    // Unwrap the constraint
    MultiTrackKinematicConstraint * internalConstraint =
      boost::any_cast< boost::shared_ptr< ::MultiTrackKinematicConstraint > > (
        cs.data().getInternal() ).get();

    // Do the fitting, this is where the real work is done
    ::RefCountedKinematicTree internalTree = fitter.fit ( internalParticles, internalConstraint );

    // Now parse the tree into the interface format
    return rave::KinematicTree ( BasicKinematicTree ( internalTree, particles ) );
  } catch ( VertexException e ) {
    edm::LogError ( "rave::KinematicTreeFactory" ) << "VertexException saying """ << e.what() << """ occured.";
    return rave::KinematicTree();
  }
}

/// TODO Allow other fitter-updator pairs
std::vector< rave::KinematicTree > KinematicTreeFactory::useParticleFitter (
  std::vector< rave::KinematicTree > trees,
  rave::KinematicConstraint cs, std::string parameters )
{
  try {
    // The fitter we are going to use
    ::KinematicParticleFitter fitter ( ParameterSetBuilder::create ( parameters ) );

    // Create the input data field
    std::vector< ::RefCountedKinematicTree > internalInputTrees;
    std::transform (
      trees.begin(),
      trees.end(),
      std::back_inserter ( internalInputTrees ),
      boost::bind ( &doTreeCasting,
                    boost::bind ( &BasicKinematicTree::getInternal,
                                  boost::bind ( &KinematicTree::data, _1 ) ) ) );

    // Unwrap the constraint
    ::KinematicConstraint * internalConstraint =
      boost::any_cast< boost::shared_ptr< ::KinematicConstraint > > (
        cs.data().getInternal() ).get();

    edm::LogInfo ( "KinematicTreeFactory" ) << "Using KinematicParticleFitter with "
    << internalInputTrees.size() << " input tree(s)";
    // Do the fitting, this is where the real work is done
    std::vector< ::RefCountedKinematicTree > internalOutputTrees =
      fitter.fit ( internalConstraint, internalInputTrees );

    // Create the output data field
    std::vector< rave::KinematicTree > outputTrees;
    for ( std::vector< ::RefCountedKinematicTree >::iterator i = internalOutputTrees.begin();
          i != internalOutputTrees.end();
          i++ )
    {
      outputTrees.push_back ( rave::KinematicTree ( rave::BasicKinematicTree ( *i ) ) );
    }

    return outputTrees;
  } catch ( VertexException e ) {
    edm::LogError ( "rave::KinematicTreeFactory" ) << "VertexException saying """ << e.what() << """ occured.";
    return std::vector< rave::KinematicTree >();
  }
}

const rave::Propagator & KinematicTreeFactory::getPropagator() const
{
  return *thePropagator;
}

const rave::MagneticField & KinematicTreeFactory::getMagneticField() const
{
  return *theField;
}

const rave::Ellipsoid3D & KinematicTreeFactory::getBeamspot() const
{
  return rave::BeamSpotSingleton::get();
}

bool KinematicTreeFactory::hasBeamspot() const
{
  return rave::BeamSpotSingleton::hasBeamSpot();
}

const int rave::KinematicTreeFactory::verbosity() const
{
  return theVerbosity;
}

}
